#!/usr/bin/env node
'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var fs = require('fs');
var commander = require('commander');
var chalk = require('chalk');
var tslib = require('tslib');
var path = require('path');
require('ts-array-extensions');
var yaml = require('yaml');
var prettier = require('prettier');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var fs__default = /*#__PURE__*/_interopDefaultLegacy(fs);
var commander__default = /*#__PURE__*/_interopDefaultLegacy(commander);
var chalk__default = /*#__PURE__*/_interopDefaultLegacy(chalk);
var path__default = /*#__PURE__*/_interopDefaultLegacy(path);
var yaml__default = /*#__PURE__*/_interopDefaultLegacy(yaml);
var prettier__default = /*#__PURE__*/_interopDefaultLegacy(prettier);

/* eslint-disable no-console */
var o = process.stdout;
var e = process.stdout;
var progressLogger = function (start) {
    var last = 0;
    return function (m) {
        o.cursorTo(start);
        o.write(m);
        if (m.length < last) {
            o.write(' '.repeat(last - m.length));
        }
        last = m.length;
    };
};
var success = function () { return chalk__default['default'].green(' -> SUCCESS'); };
var warn = function (message) {
    return chalk__default['default'].yellow(" -> WARNING: " + message);
};
var progress = function (message) {
    return chalk__default['default'].blue(" -> " + message);
};
var processLogger = function () {
    var log = function (m) {
        o.write("\n" + m);
    };
    return {
        error: function (m) {
            e.write("\n" + chalk__default['default'].red(m));
        },
        log: log,
        create: function (m) {
            log(m);
            var l = progressLogger(m.length);
            l(progress('...'));
            return l;
        }
    };
};

var mapProperties = function (properties) {
    var r = {};
    Object.keys(properties).forEach(function (key) {
        // eslint-disable-next-line @typescript-eslint/no-use-before-define
        r[key] = convertSchemaToJsonSchema(properties[key]);
    });
    return r;
};
var convertSchemaToJsonSchema = function (schema) {
    /**
     * This currently does nothing, but is a placeholder for removing
     * unsupported features from Open Api
     */
    var properties = schema.properties, rest = tslib.__rest(schema, ["properties"]);
    return tslib.__assign(tslib.__assign({}, rest), (properties ? { properties: mapProperties(properties) } : undefined));
};

var dereference = function (references, $ref, prefix) {
    if (!$ref.startsWith(prefix)) {
        throw new Error("Invalid $ref " + $ref);
    }
    var name = $ref.substring($ref.lastIndexOf('/') + 1);
    var maybe = references === null || references === void 0 ? void 0 : references[name];
    if (!maybe)
        throw new Error("Failed to deference " + $ref);
    return maybe;
};

var dereferenceParameter = function (reference, doc) {
    var $ref = reference.$ref;
    return dereference(doc.parameters, $ref, '#/parameters/');
};

var dereferenceResponseObject = function (reference, doc) {
    var $ref = reference.$ref;
    return dereference(doc.responses, $ref, '#/responses/');
};

function isReferenceObject(object) {
    if (typeof object !== 'object') {
        return false;
    }
    if (object === null) {
        return false;
    }
    return '$ref' in object;
}

var methodsOfPathItemObject = function (pathItemObject) {
    return Object.keys(pathItemObject).compactMap(function (method) {
        if (method === '$ref' || method === 'parameters') {
            return undefined;
        }
        var op = pathItemObject[method];
        if (!op) {
            return undefined;
        }
        return [method, op];
    });
};

var mapOperations = function (document) {
    return Object.keys(document.paths).flatMap(function (path) {
        var _a;
        if (typeof path !== 'string')
            return [];
        var pathObject = document.paths[path];
        var pathParameters = (_a = pathObject.parameters) !== null && _a !== void 0 ? _a : [];
        return methodsOfPathItemObject(pathObject).map(function (_a) {
            var _b = tslib.__read(_a, 2), method = _b[0], operationObject = _b[1];
            if (!('operationId' in operationObject) ||
                !operationObject.operationId) {
                throw new Error("OperationId required for " + method + " " + path);
            }
            var operationId = operationObject.operationId, _c = operationObject.parameters, parameters = _c === void 0 ? [] : _c, responses = operationObject.responses;
            var inlinedParameters = pathParameters
                .concat(parameters)
                .map(function (p) {
                return isReferenceObject(p) ? dereferenceParameter(p, document) : p;
            });
            var inlinedResponses = Object.keys(responses).compactMap(function (statusCode) {
                var response = responses[statusCode];
                if (!response)
                    return undefined;
                return {
                    statusCode: statusCode.toString(),
                    response: isReferenceObject(response)
                        ? dereferenceResponseObject(response, document)
                        : response
                };
            });
            return {
                path: path,
                method: method,
                operationId: operationId,
                operation: operationObject,
                parameters: inlinedParameters !== null && inlinedParameters !== void 0 ? inlinedParameters : [],
                responses: inlinedResponses
            };
        });
    });
};

var typeDefForEnum$1 = function (members, onwardLookup) {
    return members
        .map(function (e) {
        if (typeof e === 'string') {
            return "\"" + e + "\"";
        }
        if (e === null) {
            return 'null';
        }
        if (typeof e === 'object') {
            return onwardLookup(e);
        }
        throw new Error("Unexpected enum member");
    })
        .join(' | ');
};

var typeDefForObject$1 = function (objectSchema, recursiveLookup) {
    var required = objectSchema.required, properties = objectSchema.properties;
    if (!properties)
        return 'unknown';
    var requiredProperties = Array.isArray(required) ? required : [];
    var propertyList = Object.keys(properties).compactMap(function (propertyName) {
        if (typeof propertyName !== 'string') {
            return undefined;
        }
        var propertyObject = properties[propertyName];
        var isRequired = requiredProperties.includes(propertyName);
        var propertyType = recursiveLookup(propertyObject);
        return "" + propertyName + (isRequired ? '' : '?') + ": " + propertyType;
    });
    return "{ " + propertyList.join('; ') + " }";
};

var typeDefForReference$1 = function ($ref) {
    var name = $ref.substring($ref.lastIndexOf('/') + 1);
    return name;
};

var typeDefForSchema$1 = function (schema) {
    var _a;
    if (typeof schema === 'boolean') {
        throw new Error("Schema nodes must be an object.");
    }
    if ('$ref' in schema && schema.$ref) {
        return typeDefForReference$1(schema.$ref);
    }
    if (!('type' in schema)) {
        throw new Error("Schema nodes must have a type. " + JSON.stringify(schema));
    }
    var type = schema.type, enumProp = schema.enum, allOf = schema.allOf;
    if (enumProp) {
        return typeDefForEnum$1(enumProp, typeDefForSchema$1);
    }
    if (type === 'object') {
        var def = typeDefForObject$1(schema, typeDefForSchema$1);
        if (allOf) {
            return allOf
                .map(function (m) { return typeDefForSchema$1(m); })
                .concat(def)
                .join(' & ');
        }
        return def;
    }
    if (allOf) {
        return allOf
            .map(function (m) { return typeDefForSchema$1(m); })
            .join(' & ');
    }
    if (type === 'boolean' ||
        type === 'null' ||
        type === 'number' ||
        type === 'string') {
        return type;
    }
    if (type === 'integer') {
        return 'number';
    }
    if (type === 'array') {
        var items = schema.items;
        if (!items) {
            throw new Error("Array nodes must have an items property.");
        }
        if (Array.isArray(items)) {
            throw new Error("Array items property must be singular.");
        }
        return "Array<" + typeDefForSchema$1(items) + ">";
    }
    throw new Error("Unrecognized type " + ((_a = type === null || type === void 0 ? void 0 : type.toString()) !== null && _a !== void 0 ? _a : ''));
};

var generateResponseType = function (responses) {
    return responses
        .map(function (r) {
        var statusType = r.statusCode === 'default' ? 'number' : r.statusCode;
        var dataType = (function () {
            if (!r.response.schema) {
                return 'unknown';
            }
            if (r.response.schema.type === 'file') {
                return 'Blob';
            }
            return typeDefForSchema$1(r.response.schema);
        })();
        return "ResponseWithData<" + statusType + ", " + dataType + ">";
    })
        .join(' | ');
};

var generateMethodArguments = function (parameters) {
    if (parameters.none()) {
        return null;
    }
    var properties = parameters.map(function (v) {
        // eslint-disable-next-line @typescript-eslint/naming-convention
        var iin = v.in, name = v.name, required = v.required, rest = tslib.__rest(v, ["in", "name", "required"]);
        var typeDef = rest.type === 'file'
            ? 'Blob'
            : typeDefForSchema$1(iin === 'body' ? rest.schema : rest);
        return "" + name + (required ? '' : '?') + ": " + typeDef;
    });
    return "{ " + properties.join('; ') + " }";
};

var generateMethods = function (document, log) {
    return mapOperations(document).map(function (_a) {
        var method = _a.method, parameters = _a.parameters, path = _a.path, operationId = _a.operationId, responses = _a.responses;
        log === null || log === void 0 ? void 0 : log(progress("Adding " + method + " " + path));
        var argsType = generateMethodArguments(parameters);
        var responseType = generateResponseType(responses);
        var functionArguments = argsType ? "args: " + argsType : '';
        var bodyArg = parameters.first(function (p) { return p.in === 'body'; });
        var queryArgs = parameters.filter(function (p) { return p.in === 'query'; });
        var headerArgs = parameters.filter(function (p) { return p.in === 'header'; });
        var formDataArgs = parameters.filter(function (p) { return p.in === 'formData'; });
        var paramNames = parameters.map(function (p) { return p.name; });
        var interpolationPath = '`${this.baseUrl}' +
            path.replace(/\{(?:.*?)\}/g, function (x) { return '$' + x; }) +
            (queryArgs.any() ? '?${query}' : '') +
            '`';
        var decomposeParameters = argsType
            ? "const {" + paramNames.join(', ') + "} = args;"
            : '';
        var composeFormData = formDataArgs.none()
            ? ''
            : "\n          const formData = new FormData();\n          " + formDataArgs
                .map(function (f) { return "formData.append('" + f.name + "', " + f.name + ");"; })
                .join('\n') + "\n        ";
        var composeQuery = queryArgs.none()
            ? ''
            : "const query = qs.stringify({ " + queryArgs
                .map(function (q) { return q.name; })
                .join(', ') + " });";
        var headers = headerArgs
            .map(function (h) { return h.name; })
            .concat(bodyArg ? ["'Content-Type': 'application/json'"] : []);
        var fetchOpts = "{\n        " + tslib.__spreadArray(tslib.__spreadArray(tslib.__spreadArray([
            'method'
        ], tslib.__read((headers.any() ? ["headers: {" + headers.join(',') + "}"] : []))), tslib.__read((bodyArg ? ["body: JSON.stringify(" + bodyArg.name + ")"] : []))), tslib.__read((formDataArgs.any() ? ['body: formData'] : []))).join(',\n') + "\n      }";
        return "public async " + operationId + "(" + functionArguments + "): \n          Promise<" + responseType + "> {\n\n          " + decomposeParameters + "\n\n          " + composeQuery + "\n\n          " + composeFormData + "\n\n          const method = '" + method.toUpperCase() + "',\n          const url = " + interpolationPath + ";\n\n          const response = await fetch(url, " + fetchOpts + ");\n\n          const { status: $status } = response;\n\n          this.logger?.('REST API Call', { method, url, status: $status });\n\n          switch ($status) {\n            " + responses
            .map(function (r) {
            var _a;
            return "\n            " + (r.statusCode === 'default' ? "default" : "case " + r.statusCode) + ": return {\n              status: $status,\n              body: " + (((_a = r.response.schema) === null || _a === void 0 ? void 0 : _a.type) === 'file'
                ? 'await response.blob()'
                : "(await response.json()) as " + (r.response.schema
                    ? typeDefForSchema$1(r.response.schema)
                    : 'unknown')) + "\n            }\n            ";
        })
            .join('\n') + "\n\n            " + (responses.some(function (r) { return r.statusCode === 'default'; })
            ? ""
            : "\n                default:\n                  throw new Error(" +
                '`Unexpected status ${$status}`' +
                ")\n              ") + "\n          }\n        }";
    });
};

var generateTypes = function (document, log) {
    var _a;
    return Object.keys((_a = document.definitions) !== null && _a !== void 0 ? _a : {}).compactMap(function (key) {
        var _a;
        var value = (_a = document.definitions) === null || _a === void 0 ? void 0 : _a[key];
        if (!value) {
            return undefined;
        }
        log === null || log === void 0 ? void 0 : log(progress("adding " + key));
        var typeDef = typeDefForSchema$1(value);
        return "export type " + key + " = " + typeDef + ";";
    });
};

var fromV3$5 = function () {
    throw new Error('OpenApi 3 not supported');
};
var generateClient = function (_a) {
    var logger = _a.logger, openApiDocument = _a.openApiDocument, serviceName = _a.serviceName;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating client');
    var types = 'swagger' in openApiDocument
        ? generateTypes(openApiDocument, log)
        : fromV3$5();
    var methods = 'swagger' in openApiDocument
        ? generateMethods(openApiDocument, log)
        : fromV3$5();
    var code = "\n    import qs from 'qs';\n    import { fetch } from 'cross-fetch';\n\n    " + types.join('\n\n') + "\n\n    type ResponseWithData<TStatus, TData> = {\n      status: TStatus;\n      body: TData;\n    };\n      \n    type LogFn = (message: string, data: Record<string, unknown>) => void;\n\n    export class " + serviceName + "Client {\n      public constructor(\n        public readonly baseUrl: string,\n        private readonly logger?: LogFn\n      ) {}\n      \n      " + methods.join('\n\n') + "\n    }\n  ";
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

var typeDefForObject = function (objectSchema, recursiveLookup) {
    var required = objectSchema.required, properties = objectSchema.properties;
    if (!properties)
        return 'unknown';
    var requiredProperties = Array.isArray(required) ? required : [];
    var propertyList = Object.keys(properties).compactMap(function (propertyName) {
        if (typeof propertyName !== 'string') {
            return undefined;
        }
        var propertyObject = properties[propertyName];
        var isRequired = requiredProperties.includes(propertyName);
        var propertyType = recursiveLookup(propertyObject);
        return "" + propertyName + (isRequired ? '' : '?') + ": " + propertyType;
    });
    return "{ " + propertyList.join('; ') + " }";
};

var typeDefForEnum = function (members, onwardLookup) {
    return members
        .map(function (e) {
        if (typeof e === 'string') {
            return "\"" + e + "\"";
        }
        if (e === null) {
            return 'null';
        }
        if (typeof e === 'object') {
            if (Array.isArray(e)) {
                throw new Error("Don't know what to do with an array-like enum member");
            }
            return onwardLookup(e);
        }
        return e.toString();
    })
        .join(' | ');
};

var typeDefForReference = function ($ref) {
    var name = $ref.substring($ref.lastIndexOf('/') + 1);
    return name;
};

var typeDefForSchema = function (schema) {
    var _a;
    if (typeof schema === 'boolean') {
        throw new Error("Schema nodes must be an object.");
    }
    if ('$ref' in schema && schema.$ref) {
        return typeDefForReference(schema.$ref);
    }
    if (!('type' in schema)) {
        throw new Error("Schema nodes must have a type.");
    }
    var type = schema.type, enumProp = schema.enum, allOf = schema.allOf;
    if (enumProp) {
        return typeDefForEnum(enumProp, typeDefForSchema);
    }
    if (type === 'object') {
        var def = typeDefForObject(schema, typeDefForSchema);
        if (allOf) {
            return allOf
                .map(function (m) { return typeDefForSchema(m); })
                .concat(def)
                .join(' & ');
        }
        return def;
    }
    if (allOf) {
        return allOf.map(function (m) { return typeDefForSchema(m); }).join(' & ');
    }
    if (type === 'boolean' ||
        type === 'null' ||
        type === 'number' ||
        type === 'string') {
        return type;
    }
    if (type === 'integer') {
        return 'number';
    }
    if (type === 'array') {
        var items = schema.items;
        if (!items) {
            throw new Error("Array nodes must have an items property.");
        }
        if (Array.isArray(items)) {
            throw new Error("Array items property must be singular.");
        }
        return "Array<" + typeDefForSchema(items) + ">";
    }
    throw new Error("Unrecognized type " + ((_a = type === null || type === void 0 ? void 0 : type.toString()) !== null && _a !== void 0 ? _a : ''));
};

var initUpper = function (value) {
    return value.replace(/(?:^|\s)[a-z]/g, function (m) { return m.toUpperCase(); });
};

var parametersTypeNameTemplate = function (operationId, iin) {
    return initUpper(operationId) + "Request" + initUpper(iin === 'formData' ? 'body' : iin);
};

var responseBodyTypeNameTemplate = function (operationId, statusCode) { return "" + initUpper(operationId) + statusCode + "ResponseBody"; };

var serviceNameTemplate = function (title) {
    return title.replace(/\s+.{1}/g, function (v) { return v.slice(-1).toUpperCase(); });
};

var fromV2$3 = function (document, log) {
    return mapOperations(document).flatMap(function (_a) {
        var operationId = _a.operationId, responses = _a.responses;
        return responses.compactMap(function (_a) {
            var statusCode = _a.statusCode, response = _a.response;
            var schema = response.schema;
            if (!schema) {
                return undefined;
            }
            if (schema.type !== 'file') {
                return undefined;
            }
            var typeName = responseBodyTypeNameTemplate(operationId, statusCode);
            log === null || log === void 0 ? void 0 : log(progress("adding " + typeName));
            return "export type " + typeName + " = Buffer;";
        });
    });
};
var fromV3$4 = function () {
    throw new Error('OpenApi 3 not supported');
};
var fromJsonSchema = function (jsonSchema, log) {
    var definitions = jsonSchema.definitions;
    if (!definitions)
        return [];
    return Object.keys(definitions).compactMap(function (typeName) {
        if (typeof typeName !== 'string')
            return undefined;
        var definition = definitions[typeName];
        log === null || log === void 0 ? void 0 : log(progress(typeName));
        return "export type " + typeName + " = " + typeDefForSchema(definition) + ";";
    });
};
var generateServerTypes = function (openApiDocument, jsonSchema, log) {
    log === null || log === void 0 ? void 0 : log(progress('Generating typescript types'));
    var types = tslib.__spreadArray(tslib.__spreadArray([], tslib.__read(fromJsonSchema(jsonSchema, log))), tslib.__read(('swagger' in openApiDocument ? fromV2$3(openApiDocument, log) : fromV3$4())));
    log === null || log === void 0 ? void 0 : log(success());
    return types;
};

var initLower = function (value) {
    return value.replace(/(?:^|\s)[A-Z]/g, function (m) { return m.toLowerCase(); });
};

var generateRequestHandlersTypes = function (document, log) {
    return mapOperations(document).map(function (_a) {
        var operationId = _a.operationId, parameters = _a.parameters, responses = _a.responses;
        var responseBodyType = responses
            .compactMap(function (_a) {
            var statusCode = _a.statusCode, response = _a.response;
            var schema = response.schema;
            if (!schema) {
                return undefined;
            }
            return responseBodyTypeNameTemplate(operationId, statusCode);
        })
            .join(' | ') || 'unknown';
        var statusCodes = responses
            .map(function (_a) {
            var statusCode = _a.statusCode;
            return statusCode === 'default' ? 'number' : statusCode;
        })
            .join(' | ');
        var bodyParamType = parameters.some(function (p) { return p.in === 'body' || (p.in === 'formData' && p.type !== 'file'); })
            ? parametersTypeNameTemplate(operationId, 'body')
            : 'unknown';
        var queryParamsType = parameters.some(function (p) { return p.in === 'query'; })
            ? parametersTypeNameTemplate(operationId, 'query')
            : 'unknown';
        var pathParamsType = parameters.some(function (p) { return p.in === 'path'; })
            ? parametersTypeNameTemplate(operationId, 'path')
            : 'unknown';
        var typeName = initUpper(operationId) + "RequestHandler";
        log === null || log === void 0 ? void 0 : log(progress(typeName));
        return "\n        export type " + typeName + " = RequestHandler<\n          " + pathParamsType + ",\n          " + responseBodyType + ",\n          " + bodyParamType + ",\n          " + queryParamsType + ",\n          Record<string, any>,\n          " + statusCodes + "\n        >";
    }).concat("\n    export type RequestHandlers = {\n      " + mapOperations(document)
        .map(function (_a) {
        var operationId = _a.operationId;
        return initLower(operationId) + ": " + initUpper(operationId) + "RequestHandler";
    })
        .join(';\n') + "\n    };\n  ");
};

var fromV2$2 = function (document, log) {
    return mapOperations(document).map(function (_a) {
        var method = _a.method, path = _a.path, parameters = _a.parameters, operationId = _a.operationId;
        var handlerType = initUpper(operationId) + "RequestHandler";
        var handlerInstanceName = "" + initLower(operationId);
        var requestHandlersPath = path.replace(/\{(?:.*?)\}/g, function (x) { return ":" + x.substr(1, x.length - 2); });
        log === null || log === void 0 ? void 0 : log(progress("Adding " + method + " " + requestHandlersPath));
        var args = tslib.__spreadArray(tslib.__spreadArray([
            "'" + requestHandlersPath + "'"
        ], tslib.__read((parameters.any() ? [operationId + "Validator(options)"] : []))), [
            handlerInstanceName
        ]);
        return {
            route: "." + method + "(" + args.join(', ') + ")",
            handlerInstanceName: handlerInstanceName,
            handlerType: handlerType
        };
    });
};
var fromV3$3 = function () {
    throw new Error('OpenApi 3 not supported');
};
var generateRouter = function (openApiDocument, serviceName, log) {
    log === null || log === void 0 ? void 0 : log(progress('Router'));
    var types = 'swagger' in openApiDocument ? fromV2$2(openApiDocument, log) : fromV3$3();
    var argumentList = types
        .map(function (_a) {
        var handlerInstanceName = _a.handlerInstanceName;
        return handlerInstanceName;
    })
        .join(', ');
    return "\n    export const " + initLower(serviceName) + "Router\n      = ({ " + argumentList + " }: RequestHandlers, options?: ValidationOptions): Router => {\n      return Router()\n        " + types.map(function (t) { return t.route; }).join('\n') + "\n    }\n  ";
};

var PERMITTED_TYPES = ['body', 'path', 'query', 'header', 'formData'];
var fromV2$1 = function (document, log) {
    var _a;
    var definitions = {};
    mapOperations(document).forEach(function (_a) {
        var operationId = _a.operationId, parameters = _a.parameters, responses = _a.responses;
        var groups = parameters.groupBy(function (p) { return p.in; });
        var keys = groups.map(function (g) { return g.key; });
        if (keys.includes('body') && keys.includes('formData')) {
            throw new Error("Can not have both formData and body parameters");
        }
        groups.forEach(function (_a) {
            var key = _a.key, values = _a.values;
            if (!PERMITTED_TYPES.includes(key)) {
                throw new Error("Unsupported parameter in " + key);
            }
            if (key === 'body') {
                if (values.length > 1) {
                    throw new Error('Only a single body parameter per operation is supported');
                }
                var schema_1 = values[0].schema;
                var typeName_1 = parametersTypeNameTemplate(operationId, 'body');
                log === null || log === void 0 ? void 0 : log(progress("adding " + typeName_1));
                definitions[typeName_1] = convertSchemaToJsonSchema(schema_1);
                return;
            }
            var typeName = parametersTypeNameTemplate(operationId, key);
            var schema = {
                type: 'object',
                properties: {},
                required: []
            };
            values.forEach(function (v) {
                // eslint-disable-next-line @typescript-eslint/naming-convention
                v.in; var name = v.name, required = v.required, rest = tslib.__rest(v, ["in", "name", "required"]);
                if (rest.type === 'file') {
                    return;
                }
                schema.properties[name] = rest;
                if (required) {
                    schema.required = tslib.__spreadArray(tslib.__spreadArray([], tslib.__read(schema.required)), [name]);
                }
            });
            log === null || log === void 0 ? void 0 : log(progress("adding " + typeName));
            definitions[typeName] = convertSchemaToJsonSchema(schema);
        });
        responses.forEach(function (_a) {
            var statusCode = _a.statusCode, response = _a.response;
            var schema = response.schema;
            if (!schema) {
                return;
            }
            if (schema.type === 'file') {
                return;
            }
            var key = responseBodyTypeNameTemplate(operationId, statusCode);
            log === null || log === void 0 ? void 0 : log(progress("adding " + key));
            definitions[key] = convertSchemaToJsonSchema(schema);
        });
    });
    Object.keys((_a = document.definitions) !== null && _a !== void 0 ? _a : {}).forEach(function (key) {
        var _a;
        var value = (_a = document.definitions) === null || _a === void 0 ? void 0 : _a[key];
        if (value) {
            log === null || log === void 0 ? void 0 : log(progress("adding " + key));
            definitions[key] = convertSchemaToJsonSchema(value);
        }
    });
    var out = {
        $schema: 'http://json-schema.org/draft-07/schema#',
        definitions: definitions
    };
    return out;
};
var fromV3$2 = function () {
    throw new Error('OpenApi 3 not supported');
};
var generateJsonSchema = function (openApiDocument, log) {
    log === null || log === void 0 ? void 0 : log(progress("Generating JSON schema from Open API Document"));
    var schema = 'swagger' in openApiDocument ? fromV2$1(openApiDocument, log) : fromV3$2();
    return schema;
};

var requestPropertyName = function (type) {
    switch (type) {
        case 'header':
            return 'headers';
        case 'path':
            return 'params';
        case 'formData':
            return 'body';
        default:
            return type;
    }
};
var generateRequestValidators = function (openApiDocument, log) {
    log === null || log === void 0 ? void 0 : log(progress('Generating request body validators'));
    return mapOperations(openApiDocument).compactMap(function (_a) {
        var operationId = _a.operationId, parameters = _a.parameters;
        if (parameters.none()) {
            return undefined;
        }
        var parameterTypes = parameters
            .map(function (p) { return p.in; })
            .distinct()
            .map(function (p) {
            return [
                "validate" + initUpper(p),
                parametersTypeNameTemplate(operationId, p),
                requestPropertyName(p)
            ];
        });
        log === null || log === void 0 ? void 0 : log(progress("Generating " + operationId + "Validator"));
        return ("\n        const " + operationId + "Validator = (options?: ValidationOptions): RequestHandler => {\n          " + parameterTypes
            .map(function (_a) {
            var _b = tslib.__read(_a, 2), validator = _b[0], type = _b[1];
            return "\n            const " + validator + " = ajv.getSchema('#/definitions/" + type + "')!;\n          ";
        })
            .join('\n') + "\n\n          return (req, res, next) => {\n            if ([\n              " + parameterTypes
            .map(function (_a) {
            var _b = tslib.__read(_a, 3), validator = _b[0]; _b[1]; var requestProperty = _b[2];
            return "\n                " + validator + "(req." + requestProperty + ")\n              ";
        })
            .join(',\n') + "\n            ].every(r => r)) {\n              return next();\n            }\n\n            const errors = ([\n              " + parameterTypes
            .map(function (_a) {
            var _b = tslib.__read(_a, 3), validator = _b[0]; _b[1]; var requestProperty = _b[2];
            return "\n                [" + validator + ", '" + requestProperty + "']\n              ";
        })
            .join(',\n') + "\n            ] as const).flatMap(([validator, path]) =>\n              validator.errors?.map(e => " +
            '`${path}${e.dataPath} ${e.message}`' +
            ")\n            ).compact();\n\n            options?.logger?.(req)('Request validation failed', { errors });\n\n            res.status(400).send({\n              errors\n            });\n          };\n        };\n      ");
    });
};

var fromV3$1 = function () {
    throw new Error('OpenApi 3 not supported');
};
var generateServer = function (_a) {
    var logger = _a.logger, openApiDocument = _a.openApiDocument, serviceName = _a.serviceName;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating server');
    var jsonSchema = 'swagger' in openApiDocument
        ? generateJsonSchema(openApiDocument, log)
        : fromV3$1();
    var types = 'swagger' in openApiDocument
        ? generateServerTypes(openApiDocument, jsonSchema, log)
        : fromV3$1();
    var handlers = 'swagger' in openApiDocument
        ? generateRequestHandlersTypes(openApiDocument, log)
        : fromV3$1();
    var validators = 'swagger' in openApiDocument
        ? generateRequestValidators(openApiDocument, log)
        : fromV3$1();
    var router = 'swagger' in openApiDocument
        ? generateRouter(openApiDocument, serviceName, log)
        : fromV3$1();
    var code = "\n    import Ajv from 'ajv';\n    import { Request, Router } from 'express';\n    import { RequestHandler } from '@laurence79/express-async-request-handler';\n\n    export type ValidationOptions = {\n      logger?: (req: Request) => (message: string, data: Record<string, unknown>) => void;\n    };\n\n    " + types.join('\n\n') + "\n\n    " + handlers.join('\n\n') + "\n\n    const ajv = new Ajv({ strict: false });\n    ajv.addSchema(" + JSON.stringify(jsonSchema) + ");\n\n    " + validators.join('\n\n') + "\n\n    " + router + "\n  ";
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

/* eslint-disable prefer-template */
var fromV2 = function (document, log) {
    return mapOperations(document).map(function (_a) {
        var method = _a.method, path = _a.path, parameters = _a.parameters, operationId = _a.operationId, responses = _a.responses;
        log === null || log === void 0 ? void 0 : log(progress("Adding " + method + " " + path));
        var bodyArg = parameters.first(function (p) { return p.in === 'body' || (p.in === 'formData' && p.type !== 'file'); });
        var pathArgs = parameters.filter(function (p) { return p.in === 'path'; });
        var queryArgs = parameters.filter(function (p) { return p.in === 'query'; });
        var headerArgs = parameters.filter(function (p) { return p.in === 'header'; });
        var headersType = headerArgs.any()
            ? parametersTypeNameTemplate(operationId, 'header')
            : 'unknown';
        var pathParamsType = pathArgs.any()
            ? parametersTypeNameTemplate(operationId, 'path')
            : 'unknown';
        var queryParamsType = queryArgs.any()
            ? parametersTypeNameTemplate(operationId, 'query')
            : 'unknown';
        var requestBodyType = bodyArg
            ? parametersTypeNameTemplate(operationId, 'body')
            : 'unknown';
        var responseBodyType = responses
            .compactMap(function (_a) {
            var statusCode = _a.statusCode, response = _a.response;
            var schema = response.schema;
            if (!schema) {
                return undefined;
            }
            return responseBodyTypeNameTemplate(operationId, statusCode);
        })
            .join(' | ') || 'unknown';
        var statusCodes = responses
            .map(function (_a) {
            var statusCode = _a.statusCode;
            return statusCode === 'default' ? 'number' : statusCode;
        })
            .join(' | ');
        var methodStubType = "MethodStub<\n        " + headersType + ",\n        " + pathParamsType + ",\n        " + queryParamsType + ",\n        " + requestBodyType + ",\n        " + statusCodes + ",\n        " + responseBodyType + "\n      >";
        var requestHandlerPath = path.replace(/\{(?:.*?)\}/g, function (x) { return ":" + x.substr(1, x.length - 2); });
        return {
            operationId: operationId,
            requestHandlerPath: requestHandlerPath,
            method: method,
            methodStubType: methodStubType
        };
    });
};
var fromV3 = function () {
    throw new Error('OpenApi 3 not supported');
};
var generateStubs = function (_a) {
    var logger = _a.logger, openApiDocument = _a.openApiDocument, serviceName = _a.serviceName;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating server stubs');
    var jsonSchema = 'swagger' in openApiDocument
        ? generateJsonSchema(openApiDocument, log)
        : fromV3();
    var types = 'swagger' in openApiDocument
        ? generateServerTypes(openApiDocument, jsonSchema, log)
        : fromV3();
    var data = 'swagger' in openApiDocument ? fromV2(openApiDocument, log) : fromV3();
    var code = "\n    import { RequestHandler } from '@laurence79/express-async-request-handler';\n    import { Router } from 'express';\n\n    " + types.join('\n\n') + "\n      \n    type RequestLog<THeaders, TParams, TQuery, TBody> = {\n      headers: THeaders;\n      params: TParams;\n      query: TQuery;\n      body: TBody;\n    };\n\n    type ResponseStub<TStatus extends number, TBody> = {\n      statusCode: TStatus;\n      body: TBody;\n    };\n\n    type MethodStub<\n      THeaders,\n      TParams,\n      TQuery,\n      TRequestBody,\n      TStatus extends number,\n      TResponseBody\n    > = {\n      requestHandler: RequestHandler<\n        TParams,\n        TResponseBody,\n        TRequestBody,\n        TQuery,\n        Record<string, unknown>,\n        TStatus\n      >;\n      respondWith(data: ResponseStub<TStatus, TResponseBody>): void;\n      callLog(): RequestLog<THeaders, TParams, TQuery, TRequestBody>[];\n    };\n\n    const methodStub = <\n      THeaders,\n      TParams,\n      TQuery,\n      TRequestBody,\n      TStatus extends number,\n      TResponseBody\n    >(): MethodStub<\n      THeaders,\n      TParams,\n      TQuery,\n      TRequestBody,\n      TStatus,\n      TResponseBody\n    > => {\n      const log: RequestLog<THeaders, TParams, TQuery, TRequestBody>[] = [];\n\n      let responseStub: ResponseStub<TStatus, TResponseBody> | null = null;\n\n      return {\n        requestHandler: (req, res) => {\n          const { headers, params, query, body } = req;\n\n          log.push({\n            headers: (headers as unknown) as THeaders,\n            params,\n            query,\n            body\n          });\n\n          if (!responseStub) {\n            throw new Error('Method not stubbed');\n          }\n\n          res.status(responseStub.statusCode).send(responseStub.body);\n        },\n\n        respondWith(data) {\n          responseStub = data;\n        },\n\n        callLog() {\n          return log;\n        }\n      };\n    };\n\n    type MethodStubs = {\n      " + data
        .map(function (_a) {
        var operationId = _a.operationId, methodStubType = _a.methodStubType;
        return "readonly " + operationId + ": " + methodStubType + ";";
    })
        .join('\n') + "\n    };\n\n    export type " + serviceName + "ServiceStub = {\n      readonly middleware: RequestHandler;\n      readonly reset: () => void;\n      readonly stubs: () => MethodStubs;\n    };\n\n    export const " + initLower(serviceName) + "ServiceStub \n    = (): " + serviceName + "ServiceStub => {\n      let methodStubs: MethodStubs;\n      let currentRouter: RequestHandler;\n\n      const reset = () => {\n        methodStubs = {\n          " + data
        .map(function (_a) {
        var operationId = _a.operationId;
        return operationId + ": methodStub()";
    })
        .join(',\n') + "\n        };\n\n        currentRouter = Router()\n          " + data
        .map(function (_a) {
        var method = _a.method, operationId = _a.operationId, requestHandlerPath = _a.requestHandlerPath;
        return "." + method + "('" + requestHandlerPath + "', methodStubs." + operationId + ".requestHandler)";
    })
        .join('\n') + "\n      };\n\n      reset();\n\n      return {\n        reset,\n        middleware: (...args: Parameters<RequestHandler>) => currentRouter(...args),\n        stubs: () => methodStubs\n      };\n    };\n  ";
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

/* eslint-disable @typescript-eslint/restrict-template-expressions */
function assertNever(x) {
    throw new Error("Unexpected object: " + x);
}

var loadOpenApiDocument = function (_a) {
    var openApiDocumentFilename = _a.openApiDocumentFilename, logger = _a.logger;
    var log = logger === null || logger === void 0 ? void 0 : logger.create("Loading swagger file " + openApiDocumentFilename);
    if (!fs__default['default'].existsSync(openApiDocumentFilename)) {
        throw new Error("File " + openApiDocumentFilename + " does not exist");
    }
    var extension = path__default['default'].extname(openApiDocumentFilename);
    var data = fs__default['default'].readFileSync(openApiDocumentFilename, 'UTF-8');
    var doc = (function () {
        if (['.yaml', '.yml'].includes(extension)) {
            return yaml__default['default'].parse(data);
        }
        if (extension === '.json') {
            return JSON.parse(data);
        }
        throw new Error("Unsupported extension " + extension);
    })();
    log === null || log === void 0 ? void 0 : log(success());
    return doc;
};

var prettierStyle = {
    arrowParens: 'avoid',
    semi: true,
    trailingComma: 'none',
    singleQuote: true,
    printWidth: 80
};
var writeFile = function (_a) {
    var content = _a.content, logger = _a.logger, filename = _a.filename;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Writing file');
    log === null || log === void 0 ? void 0 : log(progress('Creating output directory'));
    fs__default['default'].mkdirSync(path__default['default'].dirname(filename), { recursive: true });
    log === null || log === void 0 ? void 0 : log(progress('Formatting content'));
    var extension = path__default['default'].extname(filename);
    var parser = (function () {
        switch (extension) {
            case '.ts':
                return 'typescript';
            case '.json':
                return 'json';
            default:
                return undefined;
        }
    })();
    var formattedContent = (function () {
        try {
            return prettier__default['default'].format(content, tslib.__assign(tslib.__assign({}, prettierStyle), { parser: parser }));
        }
        catch (e) {
            // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
            log === null || log === void 0 ? void 0 : log(warn(e.message));
            return content;
        }
    })();
    fs__default['default'].writeFileSync(filename, formattedContent);
    log === null || log === void 0 ? void 0 : log(success());
};

var defaultOptions = {
    output: 'CLIENT'
};
var generateCode = function (inputOptions) {
    var _a;
    var options = tslib.__assign(tslib.__assign({}, defaultOptions), inputOptions);
    var openApiDocument = loadOpenApiDocument(options);
    var serviceName = (_a = inputOptions.serviceName) !== null && _a !== void 0 ? _a : serviceNameTemplate(openApiDocument.info.title);
    var filename = (function () {
        var file = "" + serviceName + initUpper(options.output.toLowerCase()) + ".generated.ts";
        if (!options.outputFilename) {
            return file;
        }
        if (path__default['default'].extname(options.outputFilename)) {
            return options.outputFilename;
        }
        return path__default['default'].join(options.outputFilename, file);
    })();
    var renderDeps = tslib.__assign(tslib.__assign({}, options), { openApiDocument: openApiDocument,
        serviceName: serviceName,
        filename: filename });
    var method = (function () {
        switch (options.output) {
            case 'CLIENT':
                return generateClient;
            case 'SERVER':
                return generateServer;
            case 'STUBS':
                return generateStubs;
            default:
                return assertNever(options.output);
        }
    })();
    writeFile(tslib.__assign(tslib.__assign({}, renderDeps), { content: method(renderDeps) }));
};

var packageJsonContent = fs__default['default'].readFileSync(__dirname + "/../package.json", 'utf-8');
var packageJson = JSON.parse(packageJsonContent);
var version = packageJson.version;
var cmd = new commander__default['default'].Command('generate')
    .description('Generates typescript code from the spec.')
    .arguments('<spec-filename>')
    .option('--template <template>', 'the type to output. client, server or stubs. Default client')
    .option('--filename <filename>', 'the file to output. Defaults to <service-name><output-type>.generated.ts')
    .option('--service-name <name>', 'the name of the service. Defaults to info.title from the open api document')
    .action(function (specFilename, options) {
    var _a;
    var logger = processLogger();
    if (options.template &&
        !['client', 'server', 'stubs'].includes(options.template.toLowerCase())) {
        commander__default['default'].help({ error: true });
    }
    var output = (options.template
        ? options.template.toUpperCase()
        : 'CLIENT');
    try {
        generateCode({
            openApiDocumentFilename: specFilename,
            output: output,
            outputFilename: options.filename,
            serviceName: options.serviceName,
            logger: logger
        });
        logger.log(chalk__default['default'].green('\nCompleted successfully!\n\n'));
        process.exit(0);
    }
    catch (e) {
        logger.error("\nFAILED. " + (e instanceof Error
            ? e.message + "\n" + ((_a = e.stack) !== null && _a !== void 0 ? _a : '')
            : 'Unknown error') + "\n\n");
        process.exit(1);
    }
});
commander__default['default'].program.version(version).addCommand(cmd).parse();

exports.generateCode = generateCode;
